/*
 * lxline-i.i.c
 * Copyright (C) 2011-2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <ct.h>
#include <fd.h>
#include <it.h>
#include <l_note.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxline-drive.h>
#include <lxline-inter.h>
#include <lxline-types.h>

#define I_LIST(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_USER(t)			(*((X1f4_E4_C_USER *) (t)))
#define I_TEXT(t)			(*((X1f4_E4_C_TEXT *) (t)))

#define l_TEXT(e, output) \
    {									      \
	X1f4_E4_C_TEXT *l;						      \
									      \
	l = (void *) (output);						      \
	*l = (e);							      \
    }

#define dxline(list) \
    ((struct dxline_type *) (list))
#define lxline(list) \
    ((struct lxline_type *) (list))

typedef struct roll_type {
    struct excase_type *excase_data;
    struct lxtext_type *lxtext_data;
    unsigned miss, size;
} roll_type;

typedef struct rule_type {
    int roll;
    struct affect_type *affect_data;
    struct excase_type *excase_data;
    struct lxline_type *lxline_data;
    struct lxtext_type *lxtext_data;
    unsigned tune;
    void *pass;
} rule_type;

extern const char *const x1f4_c1_empty_string;
extern const void *const x1f4_a1_walk_link;

static int ever_list(struct lxline_type *, struct affect_type *,
		     struct dxline_type *, unsigned *);
static int rule_line(void *, void *, const char *, unsigned);
static int slip_roll(void *);

static int
ever_list(struct lxline_type *lxline_data, struct affect_type *affect_data,
	  struct dxline_type *dxline_data, unsigned *miss)
{
    int status;
    struct rule_type rule;

    rule.excase_data = lock_excase(dxline_data);
    rule.affect_data = affect_data;
    rule.lxline_data = lxline_data;
    rule.lxtext_data = lxline_data->link_f.text;

    rule.roll = 0;
    rule.tune = 0;

    status = _libx1f4i0_span_sfnote(dxline_data->line, &rule, rule_line);

    if (rule.roll) {
	*miss = rule.affect_data - affect_data;
    }

    return status;
}


static int
rule_line(void *line, void *data, const char *text, unsigned size)
{
    int link, status;

    do {
	struct affect_type *affect_data;
	struct rule_type *rule_data;

	rule_data = line;

	affect_data = rule_data->affect_data;

	if (size) {
	    struct lxline_type *lxline_data;
	    unsigned tune;

	    size++;

	    tune = rule_data->tune;
	    if (tune < size || 512 < size) {
		if (512 < size) {
		    lxline_data = rule_data->lxline_data;

		    status = lxline_data->link_v.link
			(lxline_data->link_v.data,
			 (void *) &affect_data->fxtext, size);
		    if (status) {
			status = _libx1f4i0_lxline_stat_link(lxline_data);
			if (status) {
			    break;
			}
		    } else {
			memcpy(I_TEXT(&affect_data->fxtext), text, size);
		    }
		} else {
		    lxline_data = rule_data->lxline_data;

		    status = lxline_data->link_v.link
			(lxline_data->link_v.data, &rule_data->pass, 4096);
		    if (status) {
			status = _libx1f4i0_lxline_stat_link(lxline_data);
			if (status) {
			    break;
			}
		    } else {
			char *pass;

			pass = rule_data->pass;

			memcpy(pass, text, size);

			rule_data->pass = pass + size;
			rule_data->tune = 4096 - size;

			l_TEXT(pass, &affect_data->fxtext);
		    }
		}
	    } else {
		char *pass;

		pass = rule_data->pass;

		memcpy(pass, text, size);

		rule_data->pass = pass + size;
		rule_data->tune = tune - size;

		l_TEXT(pass, &affect_data->fxtext);
	    }
	} else {
	    l_TEXT((void *) x1f4_c1_empty_string, &affect_data->fxtext);
	}

	affect_data->link = effect_link(data);
	affect_data->type = effect_type(data);

	affect_data->fxdata = *(struct fxdata_type *) (byte(data) + 8);

	link = affect_data->link;
	if (~link) {
	    struct lxtext_type *lxtext_data;

	    lxtext_data = rule_data->lxtext_data + link;

	    if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
		status = lxtext_data->lxtype.node
		    (lxtext_data->lxtype.context, (void *) &affect_data->fxdata,
		     I_USER(&affect_data->fxdata), rule_data->excase_data);
		if (status) {
		    if (size) {
			rule_data->lxline_data->link_m.free
			    (rule_data->lxline_data->link_m.data,
			     I_TEXT(&affect_data->fxtext));
		    }
		} else {
		    rule_data->roll = 1;
		    rule_data->affect_data++;
		}
	    } else {
		int (*copy)(void *, void *, void **, const void *,
			    const struct x1f4_nodelink_type *);

		copy = lxtext_data->lxtype.copy;
		if (copy) {
		    status = copy
			(lxtext_data->lxtype.context, NULL,
			 (void *) &affect_data->fxdata,
			 I_USER(&affect_data->fxdata), rule_data->excase_data);
		    if (status) {
			if (size) {
			    rule_data->lxline_data->link_m.free
				(rule_data->lxline_data->link_m.data,
				 I_USER(&affect_data->fxtext));
			}
		    } else {
			rule_data->roll = 1;
			rule_data->affect_data++;
		    }
		} else {
		    status = 0;

		    rule_data->affect_data++;
		}
	    }
	} else {
	    status = 0;

	    rule_data->affect_data++;
	}
    } while (0);

    return status;
}


static int
slip_roll(void *roll)
{
    int excess, status = 0;
    struct excase_type *excase_data;
    struct affect_type *affect_data;
    struct lxtext_type *lxtext_data;
    struct roll_type *roll_data;
    unsigned miss;

    roll_data = roll;

    affect_data = (void *) (roll_data + 1);

    excase_data = roll_data->excase_data;
    lxtext_data = roll_data->lxtext_data;

    miss = roll_data->miss;
    for (; miss; miss--) {
	int link;

	link = affect_data->link;
	if (~link) {
	    int (*free)(void *, void *, struct excase_type **,
			const struct excase_type *);
	    struct lxtext_type *lxtext_slip;

	    lxtext_slip = lxtext_data + link;

	    free = lxtext_slip->lxtype.free;
	    if (free) {
		excess = free
		    (lxtext_slip->lxtype.context, I_USER(&affect_data->fxdata),
		     NULL, excase_data);
		if (excess) {
		    status = _libx1f4i0_lxcall_land_slip(excess, status);
		}
	    }
	}

	affect_data++;
    }

    return status;
}


int
_libx1f4i0_lxline_will_list(void *lxline, void *output, void **input,
			    void **track,
			    struct x1f4_linetext_type *linetext_data,
			    unsigned count,
			    const struct x1f4_function_type *function_text,
			    int lead, int jack, int lock, void *dxline)
{
    int status;
    void *beta, **mind;

    status = lxline(lxline)->link_v.pick
	(lxline(lxline)->link_v.data, &mind, slip_roll,
	 sizeof(struct roll_type) + count * sizeof(struct affect_type), &beta);
    if (status) {
	status = _libx1f4i0_lxline_stat_link(lxline);
    } else {
	struct affect_type *affect_data;
	struct roll_type *roll_data;

	*mind = beta;

	roll_data = beta;

	affect_data = (void *) (roll_data + 1);

	roll_data->miss = 0;
	roll_data->size = count;

	roll_data->excase_data = lock_excase(dxline);
	roll_data->lxtext_data = lxline(lxline)->link_f.text;

	status = ever_list(lxline, affect_data, dxline, &roll_data->miss);
	if (status) {
	} else {
	    if (output == x1f4_a1_walk_link) {
		status = _libx1f4i0_lxrail_loop_call
		    (lxline, output, input, linetext_data, count,
		     function_text, lead, jack, lock, X1f4_E4_TEXT,
		     affect_data);
	    } else {
		status = _libx1f4i0_lxrail_form_call
		    (lxline, output, input, linetext_data, count,
		     function_text, lead, jack, lock, X1f4_E4_TEXT,
		     affect_data);
	    }
	}
    }

    return status;
}
